/*
Copyright (c) 2016-2022 Timur Gafarov

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

/**
 * Sample types and conversion utilities
 *
 * Copyright: Timur Gafarov 2016-2022.
 * License: $(LINK2 boost.org/LICENSE_1_0.txt, Boost License 1.0).
 * Authors: Timur Gafarov
 */
module dlib.audio.sample;

/**
 * dlib.audio defines four integer sample formats: signed 8-bit, signed 16-bit,
 * unsigned 8-bit, unsigned 16-bit.
 * They are for storage only. All sound processing and sample I/O
 * should be done in floating point numbers. Floating point sample is signed and
 * ranges from -1.0f to 1.0f.
 */
enum SampleFormat
{
    /// signed 8-bit
    S8,
    /// signed 16-bit
    S16,
    /// unsigned 8-bit
    U8,
    /// unsigned 16-bit
    U16
}

/**
 * Convert integer sample to floating point sample
 */
float toFloatSample(ubyte* ptr, SampleFormat format)
{
    float res;
    switch (format)
    {
        case SampleFormat.S8:
            res = *cast(byte*)ptr;
            res /= byte.max;
            break;
        case SampleFormat.S16:
            res = *cast(short*)ptr;
            res /= short.max;
            break;
        case SampleFormat.U8:
            res = *ptr;
            res /= ubyte.max;
            res = res * 2.0f - 1.0f; // normalize to range -1..1
            break;
        case SampleFormat.U16:
            res = *cast(ushort*)ptr;
            res /= ushort.max;
            res = res * 2.0f - 1.0f; // normalize to range -1..1
            break;
        default:
            break;
    }
    return res;
}

/**
 * Convert floating point sample to integer sample
 */
void fromFloatSample(ubyte* ptr, SampleFormat format, float s)
{
    switch (format)
    {
        case SampleFormat.S8:
            *cast(byte*)ptr = cast(byte)(s * byte.max);
            break;
        case SampleFormat.S16:
            *cast(short*)ptr = cast(short)(s * short.max);
            break;
        case SampleFormat.U8:
            *cast(ubyte*)ptr = cast(ubyte)((s + 1.0f) * 0.5f * ubyte.max);
            break;
        case SampleFormat.U16:
            *cast(ushort*)ptr = cast(ushort)((s + 1.0f) * 0.5f * ushort.max);
            break;
        default:
            break;
    }
}
