/*
Copyright (c) 2022 Timur Gafarov

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

/**
 * Cross-platform thread synchronization primitive
 *
 * Copyright: Timur Gafarov 2022.
 * License: $(LINK2 https://boost.org/LICENSE_1_0.txt, Boost License 1.0).
 * Authors: Timur Gafarov
 */
module dlib.core.mutex;

version(Windows)
{
    import core.sys.windows.winbase;
}
else version(Posix)
{
    import core.sys.posix.pthread;
}

/**
 * Mutex structure.
 * Uses CRITICAL_SECTION under Windows, pthread_mutex_t under Posix
 */
struct Mutex
{
    version(Windows)
    {
        CRITICAL_SECTION _mutex;
    }
    else version(Posix)
    {
        pthread_mutex_t _mutex;
    }
    
    /// Initialize the mutex
    int init()
    {
        version(Windows)
        {
            InitializeCriticalSection(&_mutex);
            return 0;
        }
        else version(Posix)
        {
            return pthread_mutex_init(&_mutex, null);
        }
        else return 0;
    }
    
    /// Enter critical section. If the mutex is already locked, block the thread until it becomes available
    int lock()
    {
        version(Windows)
        {
            EnterCriticalSection(&_mutex);
            return 0;
        }
        else version(Posix)
        {
            return pthread_mutex_lock(&_mutex);
        }
        else return 0;
    }
    
    /// Try to enter critical section. Return immediately if the mutex is already locked
    int tryLock()
    {
        version(Windows)
        {
            return !TryEnterCriticalSection(&_mutex);
        }
        else version(Posix)
        {
            return pthread_mutex_trylock(&_mutex);
        }
        else return 0;
    }
    
    /// Leave critical section
    int unlock()
    {
        version(Windows)
        {
            LeaveCriticalSection(&_mutex);
            return 0;
        }
        else version(Posix)
        {
            return pthread_mutex_unlock(&_mutex);
        }
        else return 0;
    }
    
    /// Destroy the mutex
    int destroy()
    {
        version(Windows)
        {
            DeleteCriticalSection(&_mutex);
            return 0;
        }
        else version(Posix)
        {
            return pthread_mutex_destroy(&_mutex);
        }
        else return 0;
    }
}

///
unittest
{
    import dlib.core.memory;
    import dlib.core.thread;
    
    int x = 0;
    Mutex mutex;
    mutex.init();
    
    void add()
    {
        mutex.lock();
        int local = x;
        local = local + 1;
        x = local;
        mutex.unlock();
    }
    
    void sub()
    {
        mutex.lock();
        int local = x;
        local = local - 1;
        x = local;
        mutex.unlock();
    }
    
    Thread[100] threads;
    
    foreach(i; 0..threads.length/2)
    {
        threads[i] = New!Thread(&add);
        threads[i].start();
    }
    
    foreach(i; threads.length/2..threads.length)
    {
        threads[i] = New!Thread(&sub);
        threads[i].start();
    }
    
    foreach(t; threads)
    {
        t.join();
    }
    
    assert(x == 0);
}
