/*
Copyright (c) 2014-2022 Martin Cejp

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

module dlib.filesystem.windows.file;

version(Windows)
{
    import dlib.filesystem.filesystem;
    import dlib.filesystem.windows.common;
    import dlib.core.stream;
    import dlib.core.memory;

    class WindowsFile: IOStream
    {
        HANDLE handle;
        uint accessFlags;
        bool eof = false;

        this(HANDLE handle, uint accessFlags)
        {
            this.handle = handle;
            this.accessFlags = accessFlags;
        }

        ~this()
        {
            close();
        }

        override void close()
        {
            if (handle != INVALID_HANDLE_VALUE)
            {
                CloseHandle(handle);
                handle = INVALID_HANDLE_VALUE;
            }
        }

        override bool seekable()
        {
            return true;
        }

        override StreamPos getPosition()
        {
            LONG pos_high = 0;
            LONG pos_low = SetFilePointer(handle, 0, &pos_high, FILE_CURRENT);

            wenforce(pos_low != INVALID_SET_FILE_POINTER || GetLastError() == NO_ERROR);

            return cast(StreamPos) pos_high << 32 | pos_low;
        }

        override bool setPosition(StreamPos pos)
        {
            LONG pos_high = cast(LONG)(pos >> 32);

            if (SetFilePointer(handle, cast(LONG) pos, &pos_high, FILE_BEGIN) == INVALID_SET_FILE_POINTER
                && GetLastError() != NO_ERROR)
                return false;
            else
                return true;
        }

        override StreamSize size()
        {
            DWORD size_high;
            DWORD size_low = GetFileSize(handle, &size_high);

            wenforce(size_low != INVALID_FILE_SIZE || GetLastError() == NO_ERROR);

            return cast(StreamPos) size_high << 32 | size_low;
        }

        override bool readable()
        {
            return handle != INVALID_HANDLE_VALUE && (accessFlags & FileSystem.read) && !eof;
        }

        override size_t readBytes(void* buffer, size_t count)
        {
            // TODO: make sure that count fits in a DWORD
            DWORD dwCount = cast(DWORD) count;

            DWORD dwGot = void;
            wenforce(ReadFile(handle, buffer, dwCount, &dwGot, null));

            if (dwCount > dwGot)
                eof = true;

            return dwGot;
        }

        override bool writeable()
        {
            return handle != INVALID_HANDLE_VALUE && (accessFlags & FileSystem.write);
        }

        override size_t writeBytes(const void* buffer, size_t count)
        {
            // TODO: make sure that count fits in a DWORD
            DWORD dwCount = cast(DWORD) count;

            DWORD dwGot = void;
            wenforce(WriteFile(handle, buffer, dwCount, &dwGot, null));

            return dwGot;
        }

        override void flush()
        {
        }
    }
}
