/*
Copyright (c) 2014-2022 Timur Gafarov

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

/**
 * Copyright: Timur Gafarov 2014-2022.
 * License: $(LINK2 boost.org/LICENSE_1_0.txt, Boost License 1.0).
 * Authors: Timur Gafarov, Andrey Penechko, Roman Vlasov
 */
module dlib.geometry.frustum;

import std.math;
import dlib.math.vector;
import dlib.math.matrix;
import dlib.geometry.plane;
import dlib.geometry.aabb;
import dlib.geometry.sphere;

/// Frustum object
struct Frustum
{
    union
    {
        Plane[6] planes;
        struct
        {
            Plane leftPlane;
            Plane rightPlane;
            Plane bottomPlane;
            Plane topPlane;
            Plane farPlane;
            Plane nearPlane;
        }
    }

    this(Matrix4x4f mvp)
    {
        fromMVP(mvp);
    }

    void fromMVP(Matrix4x4f mvp)
    {
        leftPlane.a = mvp[3]  + mvp[0];
        leftPlane.b = mvp[7]  + mvp[4];
        leftPlane.c = mvp[11] + mvp[8];
        leftPlane.d = mvp[15] + mvp[12];
        leftPlane.normalize();
        leftPlane.vectorof = -leftPlane.vectorof;

        rightPlane.a = mvp[3]  - mvp[0];
        rightPlane.b = mvp[7]  - mvp[4];
        rightPlane.c = mvp[11] - mvp[8];
        rightPlane.d = mvp[15] - mvp[12];
        rightPlane.normalize();
        rightPlane.vectorof = -rightPlane.vectorof;

        bottomPlane.a = mvp[3]  + mvp[1];
        bottomPlane.b = mvp[7]  + mvp[5];
        bottomPlane.c = mvp[11] + mvp[9];
        bottomPlane.d = mvp[15] + mvp[13];
        bottomPlane.normalize();
        bottomPlane.vectorof = -bottomPlane.vectorof;

        topPlane.a = mvp[3]  - mvp[1];
        topPlane.b = mvp[7]  - mvp[5];
        topPlane.c = mvp[11] - mvp[9];
        topPlane.d = mvp[15] - mvp[13];
        topPlane.normalize();
        topPlane.vectorof = -topPlane.vectorof;

        farPlane.a = mvp[3]  - mvp[2];
        farPlane.b = mvp[7]  - mvp[6];
        farPlane.c = mvp[11] - mvp[10];
        farPlane.d = mvp[15] - mvp[14];
        farPlane.normalize();
        farPlane.vectorof = -farPlane.vectorof;

        nearPlane.a = mvp[3]  + mvp[2];
        nearPlane.b = mvp[7]  + mvp[6];
        nearPlane.c = mvp[11] + mvp[10];
        nearPlane.d = mvp[15] + mvp[14];
        nearPlane.normalize();
        nearPlane.vectorof = -nearPlane.vectorof;
    }

    bool containsPoint(Vector3f point, bool checkNearPlane = false)
    {
        int res = 0;

        foreach(i, ref p; planes)
        {
            if (i == 5 && !checkNearPlane)
                break;

            if (p.distance(point) <= 0.0f)
                res++;
        }

        return (res == (checkNearPlane? 6 : 5));
    }

    bool intersectsSphere(Sphere sphere)
    {
        float d;

        foreach(i, ref p; planes)
        {
            d = p.distance(sphere.center);

            if (d > sphere.radius)
                return false;
        }

        return true;
    }

    bool intersectsAABB(
        AABB aabb,
        bool checkBoundariesOnly = false,
        bool checkNearPlane = true)
    {
        bool result = !checkBoundariesOnly; // Inside

        foreach (i, ref plane; planes)
        {
            if (i == 5 && !checkNearPlane)
                break;

            float d = dot(aabb.center, -plane.normal);

            float r = aabb.size.x * abs(plane.normal.x) +
                      aabb.size.y * abs(plane.normal.y) +
                      aabb.size.z * abs(plane.normal.z);

            float d_p_r = d + r;
            float d_m_r = d - r;

            if (d_p_r < plane.d)
            {
                result = false; // Outside
                break;
            }
            else if(d_m_r < plane.d)
                result = true; // Intersect
        }

        return result;
    }
}

///
unittest
{
    import dlib.math.transformation;
    
    Matrix4x4f mvp = perspectiveMatrix(60.0f, 16.0f/9.0f, 0.1f, 1000.0f);
    Frustum f = Frustum(mvp);
    
    assert(f.containsPoint(Vector3f(0, 0, -10)));
    
    Sphere s = Sphere(Vector3f(0, 0, 10), 1);
    assert(!f.intersectsSphere(s));
    
    AABB aabb = AABB(Vector3f(0, 0, 0), Vector3f(1, 1, 1));
    assert(f.intersectsAABB(aabb));
}
