/*
Copyright (c) 2017-2022 Timur Gafarov

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

/**
 * Animated images
 *
 * Copyright: Timur Gafarov 2017-2022.
 * License: $(LINK2 boost.org/LICENSE_1_0.txt, Boost License 1.0).
 * Authors: Timur Gafarov
 */
module dlib.image.animation;

import dlib.core.memory;
import dlib.image.image;

/**
 * Animated image interface 
 */
interface SuperAnimatedImage: SuperImage
{
    @property size_t frameSize();
    @property uint numFrames();
    @property uint currentFrame();
    @property void currentFrame(uint f);
    uint advanceFrame();
}

/**
 * Extension of standard Image that handles animation
 *
 * Description:
 * AnimatedImage can store more than one frame of pixel data.
 * Current frame can be switched with currentFrame property.
 * All usual image operations work on current frame, so
 * you can use this class with any existing dlib.image
 * functionality: save to file, apply filters, etc.
 */
class AnimatedImage(IntegerPixelFormat fmt): Image!(fmt), SuperAnimatedImage
{
    protected:
    uint _numFrames;
    uint _currentFrame = 0;
    size_t _frameSize;

    public:
    this(uint w, uint h, uint frames)
    {
        _numFrames = frames;
        super(w, h);
        _frameSize = _width * _height * _pixelSize;
    }

    override @property SuperImage dup()
    {
        auto res = new AnimatedImage!(fmt)(_width, _height, _numFrames);
        res.data[] = data[];
        return res;
    }

    override SuperImage createSameFormat(uint w, uint h)
    {
        return new AnimatedImage!(fmt)(w, h, _numFrames);
    }

    @property size_t frameSize()
    {
        return _frameSize;
    }

    @property uint numFrames()
    {
        return _numFrames;
    }

    @property uint currentFrame()
    {
        return _currentFrame;
    }

    @property void currentFrame(uint f)
    {
        if (f >= _numFrames)
            _currentFrame = _numFrames - 1;
        else
            _currentFrame = f;
    }

    uint advanceFrame()
    {
        currentFrame = currentFrame + 1;
        return _currentFrame;
    }

    override @property ubyte[] data()
    {
        if (_currentFrame >= _numFrames)
            _currentFrame = _numFrames - 1;
        size_t offset = _frameSize * _currentFrame;
        return _data[offset..offset+_frameSize];
    }

    protected override void allocateData()
    {
        _data = new ubyte[_width * _height * _pixelSize * _numFrames];
    }
}

/// Specialization of AnimatedImage for 8-bit luminance pixel format
alias AnimatedImageL8 = AnimatedImage!(IntegerPixelFormat.L8);
/// Specialization of AnimatedImage for 8-bit luminance-alpha pixel format
alias AnimatedImageLA8 = AnimatedImage!(IntegerPixelFormat.LA8);
/// Specialization of AnimatedImage for 8-bit RGB pixel format
alias AnimatedImageRGB8 = AnimatedImage!(IntegerPixelFormat.RGB8);
/// Specialization of AnimatedImage for 8-bit RGBA pixel format
alias AnimatedImageRGBA8 = AnimatedImage!(IntegerPixelFormat.RGBA8);

/// Specialization of AnimatedImage for 16-bit luminance pixel format
alias AnimatedImageL16 = AnimatedImage!(IntegerPixelFormat.L16);
/// Specialization of AnimatedImage for 16-bit luminance-alpha pixel format
alias AnimatedImageLA16 = AnimatedImage!(IntegerPixelFormat.LA16);
/// Specialization of AnimatedImage for 16-bit RGB pixel format
alias AnimatedImageRGB16 = AnimatedImage!(IntegerPixelFormat.RGB16);
/// Specialization of AnimatedImage for 16-bit RGBA pixel format
alias AnimatedImageRGBA16 = AnimatedImage!(IntegerPixelFormat.RGBA16);

/**
 * Factory class for animated images
 */
class AnimatedImageFactory: SuperImageFactory
{
    SuperImage createImage(uint w, uint h, uint channels, uint bitDepth, uint numFrames = 1)
    {
        return animatedImage(w, h, channels, bitDepth, numFrames);
    }
}

private AnimatedImageFactory _defaultAnimatedImageFactory;

/**
 * Get default AnimatedImageFactory singleton
 */
AnimatedImageFactory animatedImageFactory()
{
    if (!_defaultAnimatedImageFactory)
        _defaultAnimatedImageFactory = new AnimatedImageFactory();
    return _defaultAnimatedImageFactory;
}

/**
 * Factory function for animated images
 */
SuperImage animatedImage(uint w, uint h, uint channels, uint bitDepth, uint numFrames = 1)
in
{
    assert(channels > 0 && channels <= 4);
    assert(bitDepth == 8 || bitDepth == 16);
}
do
{
    switch(channels)
    {
        case 1:
        {
            if (bitDepth == 8)
                return new AnimatedImageL8(w, h, numFrames);
            else
                return new AnimatedImageL16(w, h, numFrames);
        }
        case 2:
        {
            if (bitDepth == 8)
                return new AnimatedImageLA8(w, h, numFrames);
            else
                return new AnimatedImageLA16(w, h, numFrames);
        }
        case 3:
        {
            if (bitDepth == 8)
                return new AnimatedImageRGB8(w, h, numFrames);
            else
                return new AnimatedImageRGB16(w, h, numFrames);
        }
        case 4:
        {
            if (bitDepth == 8)
                return new AnimatedImageRGBA8(w, h, numFrames);
            else
                return new AnimatedImageRGBA16(w, h, numFrames);
        }
        default:
            assert(0);
    }
}

/**
 * AnimatedImage that uses dlib.core.memory instead of GC
 */
class UnmanagedAnimatedImage(IntegerPixelFormat fmt): AnimatedImage!(fmt)
{
    override @property SuperImage dup()
    {
        auto res = New!(UnmanagedAnimatedImage!(fmt))(_width, _height, _numFrames);
        res.data[] = data[];
        return res;
    }

    override SuperImage createSameFormat(uint w, uint h)
    {
        return New!(UnmanagedAnimatedImage!(fmt))(w, h, _numFrames);
    }

    this(uint w, uint h, uint frames)
    {
        super(w, h, frames);
    }

    ~this()
    {
        Delete(_data);
    }

    protected override void allocateData()
    {
        _data = New!(ubyte[])(_width * _height * _pixelSize * _numFrames);
    }

    override void free()
    {
        Delete(this);
    }
}

/// Specialization of UnmanagedAnimatedImage for 8-bit luminance pixel format
alias UnmanagedAnimatedImageL8 = UnmanagedAnimatedImage!(IntegerPixelFormat.L8);
/// Specialization of UnmanagedAnimatedImage for 8-bit luminance-alpha pixel format
alias UnmanagedAnimatedImageLA8 = UnmanagedAnimatedImage!(IntegerPixelFormat.LA8);
/// Specialization of UnmanagedAnimatedImage for 8-bit RGB pixel format
alias UnmanagedAnimatedImageRGB8 = UnmanagedAnimatedImage!(IntegerPixelFormat.RGB8);
/// Specialization of UnmanagedAnimatedImage for 8-bit RGBA pixel format
alias UnmanagedAnimatedImageRGBA8 = UnmanagedAnimatedImage!(IntegerPixelFormat.RGBA8);

/// Specialization of UnmanagedAnimatedImage for 16-bit luminance pixel format
alias UnmanagedAnimatedImageL16 = UnmanagedAnimatedImage!(IntegerPixelFormat.L16);
/// Specialization of UnmanagedAnimatedImage for 16-bit luminance-alpha pixel format
alias UnmanagedAnimatedImageLA16 = UnmanagedAnimatedImage!(IntegerPixelFormat.LA16);
/// Specialization of UnmanagedAnimatedImage for 16-bit RGB pixel format
alias UnmanagedAnimatedImageRGB16 = UnmanagedAnimatedImage!(IntegerPixelFormat.RGB16);
/// Specialization of UnmanagedAnimatedImage for 16-bit RGBA pixel format
alias UnmanagedAnimatedImageRGBA16 = UnmanagedAnimatedImage!(IntegerPixelFormat.RGBA16);

/**
 * Factory class for UnmanagedAnimatedImage
 */
class UnmanagedAnimatedImageFactory: SuperImageFactory
{
    SuperImage createImage(uint w, uint h, uint channels, uint bitDepth, uint numFrames = 1)
    {
        return unmanagedAnimatedImage(w, h, channels, bitDepth, numFrames);
    }
}

/**
 * Factory function for UnmanagedAnimatedImage
 */
SuperImage unmanagedAnimatedImage(uint w, uint h, uint channels, uint bitDepth, uint numFrames = 1)
in
{
    assert(channels > 0 && channels <= 4);
    assert(bitDepth == 8 || bitDepth == 16);
}
do
{
    switch(channels)
    {
        case 1:
        {
            if (bitDepth == 8)
                return New!UnmanagedAnimatedImageL8(w, h, numFrames);
            else
                return New!UnmanagedAnimatedImageL16(w, h, numFrames);
        }
        case 2:
        {
            if (bitDepth == 8)
                return New!UnmanagedAnimatedImageLA8(w, h, numFrames);
            else
                return New!UnmanagedAnimatedImageLA16(w, h, numFrames);
        }
        case 3:
        {
            if (bitDepth == 8)
                return New!UnmanagedAnimatedImageRGB8(w, h, numFrames);
            else
                return New!UnmanagedAnimatedImageRGB16(w, h, numFrames);
        }
        case 4:
        {
            if (bitDepth == 8)
                return New!UnmanagedAnimatedImageRGBA8(w, h, numFrames);
            else
                return New!UnmanagedAnimatedImageRGBA16(w, h, numFrames);
        }
        default:
            assert(0);
    }
}
