/*
Copyright (c) 2018-2022 Oleg Baharev, Timur Gafarov

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

/**
 * Image binarization
 *
 * Copyright: Oleg Baharev, Timur Gafarov 2018-2022.
 * License: $(LINK2 boost.org/LICENSE_1_0.txt, Boost License 1.0).
 * Authors: Oleg Baharev, Timur Gafarov
 */
module dlib.image.filters.binarization;

import dlib.image.image;
import dlib.image.color;
import dlib.image.filters.histogram;

int otsuThreshold(SuperImage img)
{
    auto histogram = createHistogram(img);

    int sumOfLuminances;

    foreach (x; 0..img.width)
    foreach (y; 0..img.height)
    {
        sumOfLuminances += cast(int)(img[x, y].luminance * 255); 
    }

    auto allPixelCount = cast(double)(img.width * img.height);
    
    int bestThreshold = 0;
    int firstClassPixelCount = 0;
    int firstClassLuminanceSum = 0;
    
    double bestSigma = 0.0;

    for (int threshold = 0; threshold < 255; threshold++)
    {
        firstClassPixelCount += histogram[threshold];
        firstClassLuminanceSum += threshold * histogram[threshold];
    
        double firstClassProbability = firstClassPixelCount / allPixelCount;
        double secondClassProbability = 1.0 - firstClassProbability;

        double firstClassMean = (firstClassPixelCount == 0) ? 0 : firstClassLuminanceSum / firstClassPixelCount;
        double secondClassMean = (sumOfLuminances - firstClassLuminanceSum) / (allPixelCount - firstClassPixelCount);
        
        double meanDelta = firstClassMean - secondClassMean;
        double sigma = firstClassProbability * secondClassProbability * meanDelta * meanDelta;

        if (sigma > bestSigma)
        {
            bestSigma = sigma;
            bestThreshold = threshold;
        }
    }

    return bestThreshold;
}

/// Otsu binarization
auto otsuBinarization(SuperImage img)
{
    SuperImage res = img.createSameFormat(img.width, img.height);
    auto threshold = otsuThreshold(img);

    foreach (x; 0..img.width)
    foreach (y; 0..img.height)
    {
        auto luminance = cast(int)(img[x,y].luminance * 255);

        if (luminance > threshold)
            res[x, y] = Color4f(1.0f, 1.0f, 1.0f, 1.0f);
        else
            res[x, y] = Color4f(0.0f, 0.0f, 0.0f, 1.0f);
    }

    return res; 
}
