/*
Copyright (c) 2013-2022 Timur Gafarov

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

/**
 * 2D signal consisting of complex number data
 *
 * Copyright: Timur Gafarov 2013-2022.
 * License: $(LINK2 boost.org/LICENSE_1_0.txt, Boost License 1.0).
 * Authors: Timur Gafarov
 */
module dlib.image.signal2d;

private
{
    import dlib.core.memory;
    import dlib.math.vector;
    import dlib.math.complex;
    import dlib.math.fft;
    import dlib.image.image;
    import dlib.image.color;
}

/// Signal domain
enum SignalDomain
{
    Spatial,
    Frequency
}

/**
 * 2D signal consisting of complex number data
 */
class Signal2D
{
    Complexf[] data;
    uint width;
    uint height;
    SignalDomain domain = SignalDomain.Spatial;

    this(uint w, uint h)
    {
        width = w;
        height = h;
        data = New!(Complexf[])(width * height);
    }

    this(SuperImage img, uint channel)
    {
        width = img.width;
        height = img.height;
        data = New!(Complexf[])(width * height);

        foreach(y; 0..height)
        foreach(x; 0..width)
        {
            auto col = img[x, y];
            size_t index = (y * width + x);
            data[index].re = col.vec[channel];
            data[index].im = 0.0f;
        }
    }
    
    ~this()
    {
        Delete(data);
    }

    void fft()
    {
        foreach(y; 0..height)
        foreach(x; 0..width)
        {
            if (((x + y) & 0x1) != 0)
            {
                size_t index = (y * width + x);
                data[index].re = data[index].re * -1;
                data[index].im = data[index].im * -1;
            }
        }

        fft2(true);
        domain = SignalDomain.Frequency;
    }

    void fftInverse()
    {
        if (domain != SignalDomain.Frequency)
            return;

        fft2(false);
        domain = SignalDomain.Spatial;

        foreach(y; 0..height)
        foreach(x; 0..width)
        {
            if (((x + y) & 0x1) != 0)
            {
                size_t index = (y * width + x);
                data[index].re = data[index].re * -1;
                data[index].im = data[index].im * -1;
            }
        }
    }

    void multiply(Signal2D img, Signal2D dest)
    {
        assert(img.width == width && img.height == height && dest.width == width && dest.height == height);

        dest.domain = domain;

        foreach(i, v; data)
        {
            dest.data[i] = v * img.data[i];
        }
    }
    
    void divide(Signal2D img, Signal2D dest)
    {
        assert(img.width == width && img.height == height && dest.width == width && dest.height == height);

        dest.domain = domain;

        foreach(i, v; data)
        {
            if (img.data[i].re == 0.0f)
                dest.data[i] = v;
            else
                dest.data[i] = v / img.data[i];
        }
    }

    Complexf opIndex(int x, int y)
    {
        while(x >= width) x = width-1;
        while(y >= height) y = height-1;
        while(x < 0) x = 0;
        while(y < 0) y = 0;

        return data[y * width + x];
    }

    void fft2(bool forward)
    {
        // process rows
        Complexf[] row = New!(Complexf[])(width);
        for (int y = 0; y < height; y++)
        {
            for (int x = 0; x < width; x++)
            {
                auto index = y * width + x;
                row[x] = data[index];
            }

            fastFourierTransform(row, forward);

            for (int x = 0; x < width; x++)
            {
                auto index = y * width + x;
                data[index] = row[x];
            }
        }

        // process columns
        Complexf[] col = New!(Complexf[])(height);
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                auto index = y * width + x;
                col[y] = data[index];
            }

            fastFourierTransform(col, forward);

            for (int y = 0; y < height; y++)
            {
                auto index = y * width + x;
                data[index] = col[y];
            }
        }
        
        Delete(row);
        Delete(col);
    }
}

/// Convert Signal2D to SuperImage
void signalToImage(Signal2D chRed, Signal2D chGreen, Signal2D chBlue, SuperImage img)
{
    float maxIntensityR = 0.0f;
    float maxIntensityG = 0.0f;
    float maxIntensityB = 0.0f;
    Vector3f[] fpImage = New!(Vector3f[])(img.width * img.height);

    for(int i = 0; i < chRed.data.length; i++)
    {
        float mr, mg, mb;

        mr = chRed.data[i].magnitude;
        if (mr > maxIntensityR)
            maxIntensityR = mr;
        mg = chGreen.data[i].magnitude;
        if (mg > maxIntensityG)
            maxIntensityG = mg;
        mb = chBlue.data[i].magnitude;
        if (mb > maxIntensityB)
            maxIntensityB = mb;

        fpImage[i] = Vector3f(mr, mg, mb);
    }

    foreach(ref v; fpImage)
    {
        v.r /= maxIntensityR;
        v.g /= maxIntensityG;
        v.b /= maxIntensityB;
    }

    foreach(y; 0..img.height)
    foreach(x; 0..img.width)
    {
        size_t index = (y * img.width + x);
        Vector3f m = fpImage[index];
        img[x, y] = Color4f(m);
    }

    Delete(fpImage);
}
