/*
Copyright (c) 2016-2022 Eugene Wissner

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

/**
 * Allocator based on malloc/realloc/free
 *
 * Copyright: Eugene Wissner 2016-2022.
 * License: $(LINK2 boost.org/LICENSE_1_0.txt, Boost License 1.0).
 * Authors: Eugene Wissner
 */
module dlib.memory.mallocator;

import dlib.memory.allocator;
import core.exception;
import core.stdc.stdlib;
import std.algorithm.comparison;

/**
 * Wrapper for malloc/realloc/free from the C standard library.
 */
class Mallocator: Allocator
{
    /**
     * Allocates $(D_PARAM size) bytes of memory.
     *
     * Params:
     *     size = Amount of memory to allocate.
     *
     * Returns: The pointer to the new allocated memory.
     */
    void[] allocate(size_t size)
    {
        if (!size)
        {
            return null;
        }
        auto p = malloc(size);

        if (!p)
        {
            onOutOfMemoryError();
        }
        return p[0..size];
    }

    ///
    unittest
    {
        auto p = Mallocator.instance.allocate(20);

        assert(p.length == 20);

        Mallocator.instance.deallocate(p);
    }

    /**
     * Deallocates a memory block.
     *
     * Params:
     *     p = A pointer to the memory block to be freed.
     *
     * Returns: Whether the deallocation was successful.
     */
    bool deallocate(void[] p)
    {
        if (p !is null)
        {
            free(p.ptr);
        }
        return true;
    }

    ///
    unittest
    {
        void[] p;
        assert(Mallocator.instance.deallocate(p));

        p = Mallocator.instance.allocate(10);
        assert(Mallocator.instance.deallocate(p));
    }

    /**
     * Increases or decreases the size of a memory block.
     *
     * Params:
     *     p    = A pointer to the memory block.
     *     size = Size of the reallocated block.
     *
     * Returns: Whether the reallocation was successful.
     */
    bool reallocate(ref void[] p, size_t size)
    {
        if (!size)
        {
            deallocate(p);
            p = null;
            return true;
        }
        else if (p is null)
        {
            p = allocate(size);
            return true;
        }
        auto r = realloc(p.ptr, size);

        if (!r)
        {
            onOutOfMemoryError();
        }
        p = r[0..size];

        return true;
    }

    ///
    unittest
    {
        void[] p;

        Mallocator.instance.reallocate(p, 20);
        assert(p.length == 20);

        Mallocator.instance.reallocate(p, 30);
        assert(p.length == 30);

        Mallocator.instance.reallocate(p, 10);
        assert(p.length == 10);

        Mallocator.instance.reallocate(p, 0);
        assert(p is null);
    }

    /**
     * Returns: The alignment offered.
     */
    @property immutable(uint) alignment() const
    {
        return cast(uint) max(double.alignof, real.alignof);
    }

    /**
     * Static allocator instance and initializer.
     *
     * Returns: The global $(D_PSYMBOL Allocator) instance.
     */
    static @property Mallocator instance() @nogc nothrow
    {
        if (instance_ is null)
        {
            immutable size = __traits(classInstanceSize, Mallocator);
            void* p = malloc(size);

            if (p is null)
            {
                onOutOfMemoryError();
            }
            p[0..size] = typeid(Mallocator).initializer[];
            instance_ = cast(Mallocator) p[0..size].ptr;
        }
        return instance_;
    }

    ///
    @nogc nothrow unittest
    {
        assert(instance is instance);
    }

    private static __gshared Mallocator instance_;
}
